import json

import kubernetes
import rules
import urlman
from django.conf import settings
from django.core.cache import cache
from django.core.exceptions import ValidationError
from django.db import IntegrityError, models
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from encrypted_fields.fields import EncryptedJSONField
from kubernetes import client, config
from kubernetes.client.rest import ApiException

from servala.core import rules as perms
from servala.core.models.mixins import ServalaModelMixin
from servala.core.validators import kubernetes_name_validator


class ServiceCategory(ServalaModelMixin, models.Model):
    """
    Categories for services, e.g. "Databases", "Storage", "Compute".
    """

    name = models.CharField(max_length=100, verbose_name=_("Name"))
    description = models.TextField(blank=True, verbose_name=_("Description"))
    logo = models.ImageField(
        upload_to="public/service_categories",
        blank=True,
        null=True,
        verbose_name=_("Logo"),
    )

    parent = models.ForeignKey(
        to="self",
        on_delete=models.CASCADE,
        related_name="children",
        blank=True,
        null=True,
        verbose_name=_("Parent"),
    )

    class Meta:
        verbose_name = _("Service category")
        verbose_name_plural = _("Service categories")

    def __str__(self):
        return self.name


class Service(ServalaModelMixin, models.Model):
    """
    A service that can be offered, e.g. "PostgreSQL", "MinIO", "Kubernetes".
    """

    name = models.CharField(max_length=100, verbose_name=_("Name"))
    slug = models.SlugField(max_length=100, verbose_name=_("URL slug"), unique=True)
    category = models.ForeignKey(
        to="ServiceCategory",
        on_delete=models.PROTECT,
        related_name="services",
        verbose_name=_("Category"),
    )
    description = models.TextField(blank=True, verbose_name=_("Description"))
    logo = models.ImageField(
        upload_to="public/services", blank=True, null=True, verbose_name=_("Logo")
    )
    # TODO schema
    external_links = models.JSONField(
        null=True, blank=True, verbose_name=_("External links")
    )

    class Meta:
        verbose_name = _("Service")
        verbose_name_plural = _("Services")

    def __str__(self):
        return self.name


def validate_dict(data, required_fields=None, allow_empty=True):
    if not data:
        if allow_empty:
            return
        raise ValidationError(_("Data may not be empty!"))

    missing_fields = required_fields - set(data)
    if missing_fields:
        raise ValidationError(
            _("Missing required fields in API credentials: %(fields)s"),
            params={"fields": ", ".join(missing_fields)},
        )


def validate_api_credentials(value):
    required_fields = ("certificate-authority-data", "server", "token")
    return validate_dict(value, required_fields)


class ControlPlane(ServalaModelMixin, models.Model):
    """
    Note: ControlPlanes are called "Service Provider Zone" in the user-facing frontend.
    """

    name = models.CharField(max_length=100, verbose_name=_("Name"))
    description = models.TextField(blank=True, verbose_name=_("Description"))
    # Either contains the fields "certificate_authority_data", "server" and "token", or is empty
    api_credentials = EncryptedJSONField(
        verbose_name=_("API credentials"),
        help_text="Required fields: certificate-authority-data, server (URL), token",
        validators=[validate_api_credentials],
    )
    required_label = models.CharField(
        max_length=100,
        blank=True,
        null=True,
        verbose_name=_("Required Label"),
        help_text=_(
            "Label value for the 'appcat.vshn.io/provider-config' added to every instance on this plane."
        ),
    )

    cloud_provider = models.ForeignKey(
        to="CloudProvider",
        on_delete=models.PROTECT,
        related_name="control_planes",
        verbose_name=_("Cloud provider"),
    )

    class Meta:
        verbose_name = _("Control plane")
        verbose_name_plural = _("Control planes")

    def __str__(self):
        return self.name

    @property
    def service_definitions(self):
        return ServiceDefinition.objects.filter(
            offering_control_planes__control_plane=self
        ).distinct()

    @property
    def kubernetes_config(self):
        conf = kubernetes.client.Configuration()
        user_name = "servala-user"

        config_dict = {
            "apiVersion": "v1",
            "clusters": [
                {
                    "cluster": {
                        "certificate-authority-data": self.api_credentials[
                            "certificate-authority-data"
                        ],
                        "server": self.api_credentials["server"],
                    },
                    "name": self.name,
                },
            ],
            "contexts": [
                {
                    "context": {
                        "cluster": self.name,
                        "namespace": "default",
                        "user": user_name,
                    },
                    "name": self.name,
                }
            ],
            "current-context": self.name,
            "kind": "Config",
            "preferences": {},
            "users": [
                {
                    "name": user_name,
                    "user": {"token": self.api_credentials["token"]},
                }
            ],
        }
        config.load_kube_config_from_dict(
            config_dict=config_dict,
            client_configuration=conf,
        )
        return conf

    def get_kubernetes_client(self):
        return kubernetes.client.ApiClient(self.kubernetes_config)

    def test_connection(self):
        if not self.api_credentials:
            return False, _("No API credentials provided")

        try:
            v1 = kubernetes.client.CoreV1Api(self.get_kubernetes_client())
            namespace_count = len(v1.list_namespace().items)

            return True, _(
                "Successfully connected to Kubernetes API. Found {} namespaces."
            ).format(namespace_count)
        except ApiException as e:
            return False, _("API error: {}").format(str(e))
        except Exception as e:
            return False, _("Connection error: {}").format(str(e))

    def get_or_create_namespace(self, name):
        api_instance = kubernetes.client.CoreV1Api(self.get_kubernetes_client())
        try:
            api_instance.read_namespace(name=name)
        except kubernetes.client.ApiException as e:
            if e.status == 404:
                # Namespace does not exist, create it
                body = kubernetes.client.V1Namespace(
                    metadata=kubernetes.client.V1ObjectMeta(name=name)
                )
                api_instance.create_namespace(body=body)
            else:
                # If there's another error, raise it
                raise


class CloudProvider(ServalaModelMixin, models.Model):
    """
    A cloud provider, e.g. "Exoscale".
    """

    name = models.CharField(max_length=100, verbose_name=_("Name"))
    description = models.TextField(blank=True, verbose_name=_("Description"))
    logo = models.ImageField(
        upload_to="public/service_providers",
        blank=True,
        null=True,
        verbose_name=_("Logo"),
    )
    external_links = models.JSONField(
        null=True, blank=True, verbose_name=_("External links")
    )

    class Meta:
        verbose_name = _("Cloud provider")
        verbose_name_plural = _("Cloud providers")

    def __str__(self):
        return self.name


class Plan(ServalaModelMixin, models.Model):
    """
    Each service offering can have multiple plans, e.g. for different tiers.
    """

    name = models.CharField(max_length=100, verbose_name=_("Name"))
    description = models.TextField(blank=True, verbose_name=_("Description"))
    # TODO schema
    features = models.JSONField(verbose_name=_("Features"), null=True, blank=True)
    # TODO schema
    pricing = models.JSONField(verbose_name=_("Pricing"), null=True, blank=True)
    term = models.PositiveIntegerField(
        verbose_name=_("Term"), help_text=_("Term in months")
    )
    service_offering = models.ForeignKey(
        to="ServiceOffering",
        on_delete=models.PROTECT,
        related_name="plans",
        verbose_name=_("Service offering"),
    )

    class Meta:
        verbose_name = _("Plan")
        verbose_name_plural = _("Plans")

    def __str__(self):
        return self.name


def validate_api_definition(value):
    required_fields = ("group", "version", "kind")
    return validate_dict(value, required_fields)


class ServiceDefinition(ServalaModelMixin, models.Model):
    """
    Configuration/service implementation: contains information on which
    CompositeResourceDefinition (aka XRD) implements a service on a ControlPlane.

    Is required in order to query the OpenAPI spec for dynamic form generation.
    """

    name = models.CharField(max_length=100, verbose_name=_("Name"))
    description = models.TextField(blank=True, verbose_name=_("Description"))
    api_definition = models.JSONField(
        verbose_name=_("API Definition"),
        help_text=_("Contains group, version, and kind information"),
        null=True,
        blank=True,
    )
    service = models.ForeignKey(
        to="Service",
        on_delete=models.CASCADE,
        related_name="service_definitions",
        verbose_name=_("Service"),
    )

    @property
    def control_planes(self):
        return ControlPlane.objects.filter(
            offering_connections__service_definition=self
        ).distinct()

    @property
    def service_offerings(self):
        return ServiceOffering.objects.filter(
            control_plane_connections__service_definition=self
        ).distinct()

    class Meta:
        verbose_name = _("Service definition")
        verbose_name_plural = _("Service definitions")

    def __str__(self):
        return self.name


class ControlPlaneCRD(ServalaModelMixin, models.Model):
    """
    Each combination of ServiceOffering and ControlPlane can have a different
    ServiceDefinition, which is here modeled as basically a "through" model.
    """

    service_offering = models.ForeignKey(
        to="ServiceOffering",
        on_delete=models.CASCADE,
        related_name="control_plane_connections",
        verbose_name=_("Service offering"),
    )
    control_plane = models.ForeignKey(
        to="ControlPlane",
        on_delete=models.CASCADE,
        related_name="offering_connections",
        verbose_name=_("Control plane"),
    )
    service_definition = models.ForeignKey(
        to="ServiceDefinition",
        on_delete=models.PROTECT,
        related_name="offering_control_planes",
        verbose_name=_("Service definition"),
    )

    class Meta:
        verbose_name = _("ControlPlane CRD")
        verbose_name_plural = _("ControlPlane CRDs")
        unique_together = [("service_offering", "control_plane")]

    def __str__(self):
        return f"{self.service_offering} on {self.control_plane} with {self.service_definition}"

    @cached_property
    def group(self):
        return self.service_definition.api_definition["group"]

    @cached_property
    def version(self):
        return self.service_definition.api_definition["version"]

    @cached_property
    def kind(self):
        return self.service_definition.api_definition["kind"]

    @cached_property
    def kind_plural(self):
        return self.resource_definition.status.accepted_names.plural

    @cached_property
    def resource_definition(self):
        client = self.control_plane.get_kubernetes_client()

        extensions_api = kubernetes.client.ApiextensionsV1Api(client)
        crds = extensions_api.list_custom_resource_definition()
        matching_crd = None
        for crd in crds.items:
            if matching_crd:
                break
            if crd.spec.group == self.group:
                for served_version in crd.spec.versions:
                    if served_version.name == self.version and served_version.served:
                        if crd.spec.names.kind == self.kind:
                            matching_crd = crd
                            break
        return matching_crd

    @cached_property
    def resource_schema(self):
        cache_key = f"servala:crd:schema:{self.pk}"
        if result := cache.get(cache_key):
            return result

        for v in self.resource_definition.spec.versions:
            if v.name == self.version:
                result = v.schema.open_apiv3_schema.to_dict()
                timeout_seconds = 60 * 60 * 24
                cache.set(cache_key, result, timeout=timeout_seconds)
                return result

    @cached_property
    def django_model(self):
        from servala.core.crd import generate_django_model

        kwargs = {
            "group": self.group,
            "version": self.version,
            "kind": self.kind,
        }
        return generate_django_model(self.resource_schema, **kwargs)

    @cached_property
    def model_form_class(self):
        from servala.core.crd import generate_model_form_class

        return generate_model_form_class(self.django_model)


class ServiceOffering(ServalaModelMixin, models.Model):
    """
    A service offering, e.g. "PostgreSQL on AWS", "MinIO on GCP".
    """

    service = models.ForeignKey(
        to="Service",
        on_delete=models.PROTECT,
        related_name="offerings",
        verbose_name=_("Service"),
    )
    provider = models.ForeignKey(
        to="CloudProvider",
        on_delete=models.PROTECT,
        related_name="offerings",
        verbose_name=_("Provider"),
    )
    description = models.TextField(blank=True, verbose_name=_("Description"))

    class Meta:
        verbose_name = _("Service offering")
        verbose_name_plural = _("Service offerings")

    def __str__(self):
        return _("{service_name} at {provider_name}").format(
            service_name=self.service.name, provider_name=self.provider.name
        )

    @property
    def control_planes(self):
        return ControlPlane.objects.filter(
            offering_connections__service_offering=self
        ).distinct()


class ServiceInstance(ServalaModelMixin, models.Model):
    """
    The source of truth for service instances is Kubernetes.
    The Django model only contains metadata, all other information is queried
    on the fly.
    """

    name = models.CharField(
        max_length=63, verbose_name=_("Name"), validators=[kubernetes_name_validator]
    )
    organization = models.ForeignKey(
        to="core.Organization",
        on_delete=models.PROTECT,
        verbose_name=_("Organization"),
        related_name="service_instances",
    )
    created_by = models.ForeignKey(
        to="core.User",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name="+",
    )
    context = models.ForeignKey(
        to="core.ControlPlaneCRD",
        related_name="service_instances",
        on_delete=models.PROTECT,
    )

    is_deleted = models.BooleanField(default=False)
    deleted_at = models.DateTimeField(null=True, blank=True)
    deleted_by = models.ForeignKey(
        to="core.User",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name="+",
    )

    class Meta:
        verbose_name = _("Service instance")
        verbose_name_plural = _("Service instances")
        # Names are unique per de-facto namespace, which is defined by the
        # Organization + ServiceDefinition (group, version) + the ControlPlane.
        unique_together = [("name", "organization", "context")]
        rules_permissions = {
            "view": rules.is_staff | perms.is_organization_member,
            "change": rules.is_staff | perms.is_organization_member,
            "delete": rules.is_staff | perms.is_organization_admin,
            "add": rules.is_authenticated,
        }

    class urls(urlman.Urls):
        base = "{self.organization.urls.instances}{self.name}/"

    @classmethod
    def create_instance(cls, name, organization, context, created_by, spec_data):
        # Ensure the namespace exists
        context.control_plane.get_or_create_namespace(organization.namespace)
        try:
            instance = cls.objects.create(
                name=name,
                organization=organization,
                created_by=created_by,
                context=context,
            )
        except IntegrityError:
            raise ValidationError(
                _(
                    "An instance with this name already exists in this organization. Please choose a different name."
                )
            )

        try:
            create_data = {
                "apiVersion": f"{context.group}/{context.version}",
                "kind": context.kind,
                "metadata": {
                    "name": name,
                    "namespace": organization.namespace,
                },
                "spec": spec_data or {},
            }
            if label := context.control_plane.required_label:
                create_data["metadata"]["labels"] = {settings.DEFAULT_LABEL_KEY: label}
            api_instance = client.CustomObjectsApi(
                context.control_plane.get_kubernetes_client()
            )
            api_instance.create_namespaced_custom_object(
                group=context.group,
                version=context.version,
                namespace=organization.namespace,
                plural=context.kind_plural,
                body=create_data,
            )
        except Exception as e:
            instance.delete()
            if isinstance(e, ApiException):
                try:
                    error_body = json.loads(e.body)
                    reason = error_body.get("message", str(e))
                    raise ValidationError(_("Kubernetes API error: {}").format(reason))
                except (ValueError, TypeError):
                    raise ValidationError(_("Kubernetes API error: {}").format(str(e)))
            raise ValidationError(_("Error creating instance: {}").format(str(e)))
        return instance

    @cached_property
    def kubernetes_object(self):
        """Fetch the Kubernetes custom resource object"""
        try:
            api_instance = client.CustomObjectsApi(
                self.context.control_plane.get_kubernetes_client()
            )

            return api_instance.get_namespaced_custom_object(
                group=self.context.group,
                version=self.context.version,
                namespace=self.organization.namespace,
                plural=self.context.kind_plural,
                name=self.name,
            )
        except ApiException as e:
            if e.status == 404:
                return None
            raise

    @cached_property
    def spec(self):
        if not self.kubernetes_object:
            return {}
        if not (spec := self.kubernetes_object.get("spec")):
            return {}

        # Remove fields that shouldn't be displayed
        spec = spec.copy()
        spec.pop("resourceRef", None)
        spec.pop("writeConnectionSecretToRef", None)
        return spec

    @cached_property
    def status_conditions(self):
        if not self.kubernetes_object:
            return []
        if not (status := self.kubernetes_object.get("status")):
            return []
        return status.get("conditions") or []

    @cached_property
    def connection_credentials(self):
        """
        Get connection credentials directly from the resource's writeConnectionSecretToRef
        after checking that secret conditions are available.
        """
        if not self.kubernetes_object:
            return {}

        # Check if secrets are available based on conditions
        secrets_available = any(
            [
                condition.get("type") == "Ready" and condition.get("status") == "True"
                for condition in self.status_conditions
            ]
        )
        if not secrets_available:
            return {}

        spec = self.kubernetes_object.get("spec")
        if not (secret_ref := spec.get("writeConnectionSecretToRef")):
            return {}
        if not (secret_name := secret_ref.get("name")):
            return {}

        try:
            # Get the secret data
            v1 = kubernetes.client.CoreV1Api(
                self.context.control_plane.get_kubernetes_client()
            )
            secret = v1.read_namespaced_secret(
                name=secret_name, namespace=self.organization.namespace
            )

            # Secret data is base64 encoded
            credentials = {}
            if hasattr(secret, "data") and secret.data:
                import base64

                for key, value in secret.data.items():
                    try:
                        credentials[key] = base64.b64decode(value).decode("utf-8")
                    except Exception:
                        credentials[key] = f"<binary data: {len(value)} bytes>"

            return credentials
        except ApiException as e:
            return {"error": str(e)}
        except Exception as e:
            return {"error": str(e)}
